// ignore_for_file: non_constant_identifier_names

import 'dart:async';

import 'package:galaxeus_lib/galaxeus_lib.dart';
import "package:telegram_client/telegram_client/extension/extension.dart";
// import 'package:path/path.dart';
import 'package:telegram_client/tdlib/tdlib_ffi/tdlib.dart';
import 'package:telegram_client/telegram_bot_api/update_bot.dart';
import 'package:telegram_client/telegram_client.dart';
import "package:telegram_client/telegram_client/scheme/scheme.dart"
    as telegram_client_scheme;
import 'package:telegram_client/telegram_client/telegram_client_type.dart';
// import 'package:telegram_client/telegram_client/type_def_invoke.dart';

class TelegramClientTdlibOption {
  Map<dynamic, dynamic>? clientOption;
  bool is_cli;
  int? clientId;
  Duration? invokeTimeOut;

  Duration? delayUpdate;
  double timeOutUpdate;
  EventEmitter? eventEmitter;
  bool isAutoGetChat;

  FutureOr<Map<dynamic, dynamic>> Function(String, int, LibTdJson)?
      on_get_invoke_data;
  FutureOr<void> Function(dynamic, LibTdJson)? on_receive_update;
  FutureOr<String> Function(int, LibTdJson)? on_generate_extra_invoke;
  bool isInvokeThrowOnError;
  Duration delayInvoke;
  TelegramClientTdlibOption({
    required this.isAutoGetChat,
    this.clientOption,
    required this.is_cli,
    this.clientId,
    this.invokeTimeOut,
    required this.timeOutUpdate,
    required this.delayInvoke,
    required this.delayUpdate,
    this.on_generate_extra_invoke,
    this.on_get_invoke_data,
    this.eventEmitter,
    this.on_receive_update,
    required this.isInvokeThrowOnError,
  });
}

class TelegramClientTelegramBotApiOption {
  Map clientOption;

  TelegramClientTelegramBotApiOption({
    required this.clientOption,
  });
}

/// return original data json
class TelegramClientLib {
  late Tdlib tdlib;
  late TelegramBotApi telegramBotApi;
  EventEmitter event_emitter = EventEmitter();

  String event_update;
  String event_invoke;

  /// return original data json
  TelegramClientLib({
    EventEmitter? eventEmitter,
    this.event_invoke = "tg_client_invoke",
    this.event_update = "tg_client_update",
  }) {
    if (eventEmitter != null) {
      event_emitter = eventEmitter;
    }
  }

  /// return original data json
  Future<void> init({
    String? pathTdlib,
    required TelegramClientTdlibOption telegramClientTdlibOption,
    required TelegramClientTelegramBotApiOption
        telegramClientTelegramBotApiOption,
  }) async {
    tdlib = Tdlib(
      pathTdl: pathTdlib,
      clientOption: telegramClientTdlibOption.clientOption,
      clientId: telegramClientTdlibOption.clientId,
      invokeTimeOut: telegramClientTdlibOption.invokeTimeOut,
      event_invoke: event_invoke,
      event_update: event_update,
      delayUpdate: telegramClientTdlibOption.delayUpdate,
      delayInvoke: telegramClientTdlibOption.delayInvoke,
      isAutoGetChat: telegramClientTdlibOption.isAutoGetChat,
      on_generate_extra_invoke:
          telegramClientTdlibOption.on_generate_extra_invoke,
      on_get_invoke_data: telegramClientTdlibOption.on_get_invoke_data,
      on_receive_update: telegramClientTdlibOption.on_receive_update,
      isInvokeThrowOnError: telegramClientTdlibOption.isInvokeThrowOnError,
      eventEmitter: event_emitter,
    );
    telegramBotApi = TelegramBotApi(
      tokenBot: "",
      clientOption: telegramClientTelegramBotApiOption.clientOption,
      eventEmitter: event_emitter,
      alfred: null,
    );
  }
}

/// return original data json
class TelegramClientData {
  int tdlib_client_id;
  TelegramClientType telegramClientType;
  String telegram_bot_api_token_bot;

  /// return original data json
  TelegramClientData({
    required this.telegram_bot_api_token_bot,
    required this.telegramClientType,
    required this.tdlib_client_id,
  });

  /// return original data json
  static TelegramClientData tdlib({
    required int tdlib_client_id,
  }) {
    return TelegramClientData(
      telegram_bot_api_token_bot: "",
      telegramClientType: TelegramClientType.tdlib,
      tdlib_client_id: tdlib_client_id,
    );
  }

  /// return original data json
  static TelegramClientData telegramBotApi({
    required String token_bot,
  }) {
    return TelegramClientData(
      telegram_bot_api_token_bot: token_bot,
      telegramClientType: TelegramClientType.telegam_bot_api,
      tdlib_client_id: 0,
    );
  }
}

class UpdateTelegramClient {
  Map rawData;
  Map query;
  Map client_option;
  TelegramClientData telegramClientData;
  UpdateTelegramClient({
    required this.rawData,
    required this.query,
    required this.client_option,
    required this.telegramClientData,
  });
}

/// return original data json
class TelegramClient {
  late TelegramClientLib telegram_client_lib;
  // late TelegramClientData telegram_client_data;

  /// return original data json
  TelegramClient({
    required this.telegram_client_lib,
    // required TelegramClientData telegramClientData,
  }) {
    // telegram_client_lib.telegramBotApi.on(telegram_client_lib.telegramBotApi.event_update, (UpdateBot updateBot) async {
    //   try {
    //     eventEmitter.emit(telegram_client_lib.event_update, null, updateBot);
    //     // ignore: empty_catches
    //   } catch (e) {}
    // });
    // telegram_client_lib.tdlib.on(telegram_client_lib.tdlib.event_update, (UpdateTd updateTd) async {
    //   // print(updateTd.raw);
    //   try {
    //     eventEmitter.emit(telegram_client_lib.event_update, null, updateTd);
    //     // ignore: empty_catches
    //   } catch (e) {}
    // });
  }

  Listener on({
    required String event_name,
    required FutureOr<dynamic> Function(
            UpdateTelegramClient updateTelegramClient)
        onUpdate,
    required FutureOr<dynamic> Function() onError,
  }) {
    return telegram_client_lib.event_emitter.on(event_name, null,
        (ev, context) async {
      try {
        if (ev.eventData is UpdateBot) {
          UpdateBot updateBot = (ev.eventData as UpdateBot);
          await onUpdate(
            UpdateTelegramClient(
              rawData: updateBot.body,
              query: updateBot.query,
              client_option: {},
              telegramClientData: TelegramClientData(
                tdlib_client_id: 0,
                telegram_bot_api_token_bot: "",
                telegramClientType: TelegramClientType.telegam_bot_api,
              ),
            ),
          );
        }

        if (ev.eventData is TdlibIsolateReceiveData) {
          TdlibIsolateReceiveData tdlibIsolateReceiveData =
              (ev.eventData as TdlibIsolateReceiveData);

          UpdateTd updateTd = UpdateTd(
            update: tdlibIsolateReceiveData.updateData,
            client_id: tdlibIsolateReceiveData.clientId,
            client_option: telegram_client_lib
                .tdlib.clients[tdlibIsolateReceiveData.clientId]!.client_option,
          );
          await onUpdate(
            UpdateTelegramClient(
              rawData: updateTd.raw,
              query: {},
              client_option: updateTd.client_option,
              telegramClientData: TelegramClientData(
                tdlib_client_id: updateTd.client_id,
                telegram_bot_api_token_bot: "",
                telegramClientType: TelegramClientType.tdlib,
              ),
            ),
          );
        }
        // ignore: empty_catches
      } catch (e) {}
    });
  }

  /// return original data json
  FutureOr<telegram_client_scheme.JsonScheme> invoke({
    required Map parameters,
    required TelegramClientData telegramClientData,
    bool is_form = false,
    String? urlApi,
    String? clientType,
    void Function(int bytes, int bytess)? onUploadProgress,
    bool isVoid = false,
    Duration? delayDuration,
    Duration? invokeTimeOut,
    String? extra,
    bool? isAutoGetChat,
    bool isInvokeThrowOnError = true,
    FutureOr<String> Function(int client_id, LibTdJson libTdJson)?
        onGenerateExtraInvoke,
    FutureOr<Map<dynamic, dynamic>> Function(
            String, int client_id, LibTdJson libTdJson)?
        onGetInvokeData,
  }) async {
    // telegramClientLib ??= telegram_client_lib;
    if (telegramClientData.telegramClientType ==
        TelegramClientType.telegam_bot_api) {
      Map respond = await telegram_client_lib.telegramBotApi.invoke(
        parameters["@type"],
        tokenBot: telegramClientData.telegram_bot_api_token_bot,
        parameters: parameters,
        is_form: is_form,
        urlApi: urlApi,
        clientType: clientType,
        isThrowOnError: isInvokeThrowOnError,
        onUploadProgress: onUploadProgress,
      );
      if (respond["@type"] == "error") {
        return telegram_client_scheme.Error(respond);
      }
      return telegram_client_scheme.JsonScheme(respond);
    }
    if ((telegramClientData.telegramClientType == TelegramClientType.tdlib)) {
      Map respond = (await telegram_client_lib.tdlib.invoke(
        parameters["@type"],
        parameters: parameters,
        clientId: telegramClientData.tdlib_client_id,
        isVoid: isVoid,
        delayDuration: delayDuration,
        invokeTimeOut: invokeTimeOut,
        extra: extra,
        isAutoGetChat: isAutoGetChat,
        isInvokeThrowOnError: isInvokeThrowOnError,
        onGenerateExtraInvoke: onGenerateExtraInvoke,
        onGetInvokeData: onGetInvokeData,
      ));
      if (respond["@type"] == "error") {
        return telegram_client_scheme.Error(respond);
      }
      return telegram_client_scheme.JsonScheme(respond);
    }
    telegram_client_scheme.Error error = telegram_client_scheme.Error.create(
      special_extra: "error",
      code: 500,
      message: "telegram_client_type_not_found",
      description: "",
    );
    if (isInvokeThrowOnError == true) {
      throw error;
    }
    return error;
  }

  /// return original data json
  FutureOr<telegram_client_scheme.JsonScheme> request({
    required Map parameters,
    required TelegramClientData telegramClientData,
    bool is_form = false,
    String? urlApi,
    String? clientType,
    void Function(int bytes, int bytess)? onUploadProgress,
    bool isVoid = false,
    Duration? delayDuration,
    Duration? invokeTimeOut,
    String? extra,
    bool? isAutoGetChat,
    bool isInvokeThrowOnError = true,
    FutureOr<String> Function(int client_id, LibTdJson libTdJson)?
        onGenerateExtraInvoke,
    FutureOr<Map<dynamic, dynamic>> Function(
            String, int client_id, LibTdJson libTdJson)?
        onGetInvokeData,
  }) async {
    if (telegramClientData.telegramClientType != TelegramClientType.tdlib) {
      return await invoke(
        parameters: parameters,
        telegramClientData: telegramClientData,
        is_form: is_form,
        urlApi: urlApi,
        clientType: clientType,
        onUploadProgress: onUploadProgress,
        isVoid: isVoid,
        delayDuration: delayDuration,
        invokeTimeOut: invokeTimeOut,
        extra: extra,
        isAutoGetChat: isAutoGetChat,
        isInvokeThrowOnError: isInvokeThrowOnError,
        onGenerateExtraInvoke: onGenerateExtraInvoke,
        onGetInvokeData: onGetInvokeData,
      );
    }
    String method = (parameters["@type"] as String);

    /// return original data json
    FutureOr<telegram_client_scheme.JsonScheme> callApiInvoke({
      required Map parameters,
    }) async {
      return await invoke(
        parameters: parameters,
        telegramClientData: telegramClientData,
        is_form: is_form,
        urlApi: urlApi,
        clientType: clientType,
        onUploadProgress: onUploadProgress,
        isVoid: isVoid,
        delayDuration: delayDuration,
        invokeTimeOut: invokeTimeOut,
        extra: extra,
        isAutoGetChat: isAutoGetChat,
        isInvokeThrowOnError: isInvokeThrowOnError,
        onGenerateExtraInvoke: onGenerateExtraInvoke,
        onGetInvokeData: onGetInvokeData,
      );
    }

    if (RegExp(r"^(getChat)$", caseSensitive: false).hasMatch(method)) {
      return await getChat(
        parameters: parameters,
        callApiInvoke: callApiInvoke,
      );
    }
    if (RegExp(r"^(getUser)$", caseSensitive: false).hasMatch(method)) {
      return await getUser(
        parameters: parameters,
        callApiInvoke: callApiInvoke,
      );
    }

    telegram_client_scheme.Error error = telegram_client_scheme.Error.create(
      special_extra: "error",
      code: 500,
      message: "telegram_client_type_not_found",
      description: "",
    );
    if (isInvokeThrowOnError == true) {
      throw error;
    }
    return error;
  }
}
